#include "btserverdefines.h"
#include "btclient.h"
#include "btttmp.h"
#include "btutils.h"

#define INVALIDATE_STATECACHE( handle ) stateCache[ handle ] = -1;
#define INVALIDATE_ERRORCACHE( handle ) errorCache[ handle ] = -1;

//PLEASECHECK (Victor): REMOVE!
#include <time.h>
#include <sys/time.h>
TTicks OSALTickCount( void )
{
    timeval currentTime;
    gettimeofday(&currentTime, NULL);
    TTicks res = 1000;
    res *= currentTime.tv_sec;

    return res + (currentTime.tv_usec/1000);
} // OSALTickCount

#define BTC_NEW_INTERFACE

int RemapState(int state)
{
#ifdef BTC_NEW_INTERFACE
    if (state == -1)
        return state;
    switch(state) {
    case BT_STATE_INITIALIZED:
    case BT_STATE_START_LISTENING:
    case BT_STATE_LISTENING:
    case BT_STATE_APP_CLOSING:
        return IBtClient::BTC_LISTENING;
    case BT_STATE_START_CONNECTING:
    case BT_STATE_REMOVE_CONNECTION:
    case BT_STATE_RETRIEVING_CHANNEL:
    case BT_STATE_RETRIEVE_CHANNEL:
    case BT_STATE_CONNECTING:
    case BT_STATE_CHECK_PAIRING:
        return IBtClient::BTC_OPENING;
    case BT_STATE_CONNECTED:
        return IBtClient::BTC_OPEN;
    case BT_STATE_APP_START:
    case BT_STATE_APP_REMOVE_CONNECTION:
        return IBtClient::BTC_OPEN_SETUP;
    case BT_STATE_APP_READY:
        return IBtClient::BTC_OPEN_READY;
    case BT_STATE_CLOSING:
        return IBtClient::BTC_CLOSING;
    default:
        return IBtClient::BTC_CLOSED;
    }
#else
    return state;
#endif
}

int RemapError(int error)
{
#ifdef BTC_NEW_INTERFACE
    switch(error) {
    case BT_ERR_NONE:
        return IBtClient::ERR_OK;
    case BT_ERR_CONNECTION:
        return IBtClient::ERR_CONNECTION;
    case BT_ERR_STATE:
        return IBtClient::ERR_STATE;
    default:
        return IBtClient::ERR_FAILED;
    }
#else
    return error;
#endif
}

int RemapProfile(int profile)
{
#ifdef BTC_NEW_INTERFACE
    switch(profile) {
    case IBtClient::PROF_HF:
        return BT_PROFILE_HF;
    case IBtClient::PROF_DUN:
        return BT_PROFILE_DUN;
    case IBtClient::PROF_HS:
        return BT_PROFILE_HS;
    case IBtClient::PROF_SP:
        return BT_PROFILE_SP;
    case IBtClient::PROF_AVRCP_CONTROL:
        return BT_PROFILE_AVRCP;
    default:
        return BT_PROFILE_UNDEFINED;
    }
#else
    return profile;
#endif
}

//this will be returned if communication with the server fails
#define TTMP_ERROR ERR_FAILED

BtClient::BtClient() :
  iConnection(NULL),
  iMessage(NULL)
{
	for ( int i = 0; i < BT_MAX_HANDLES; i++ ) {
		stateCache[ i ] = -1;
		errorCache[ i ] = -1;
		receiveCache[ i ] = -1;
	}
}

BtClient::~BtClient()
{
    Free();
}
int BtClient::Construct()
{
	int res = TTMP_ERR_NONE;
	DEBUG_LOG1("Connecting to the server on %s...", SERVER_SOCKET_ADDR);
	res = TTMPConnectionCreate(&iConnection, BT_TTMP_TIMEOUT, BTCL_VER_MAJOR, BTCL_VER_MINOR);
	if (res < TTMP_ERR_NONE) {
		DEBUG_LOG("Can not create a connection object");
		return ERR_FAILED;
	}
	res = TTMPConnectionConnect(iConnection, SERVER_SOCKET_ADDR);
	if (res < TTMP_ERR_NONE) {
		DEBUG_LOG1("Can not connect to the server %s", SERVER_SOCKET_ADDR);
		TTMPConnectionDestroy(iConnection);
    		iConnection = NULL;
		return ERR_FAILED;
	}
	res = TTMPMessageCreate(&iMessage);
	if (res < TTMP_ERR_NONE) {
		DEBUG_LOG("Unable to create message");
		TTMPMessageDestroy(iMessage);
    		iMessage = NULL;
		TTMPConnectionDestroy(iConnection);
    		iConnection = NULL;
		return ERR_FAILED;
	}
	iCmdId = 0;
	DEBUG_LOG("Connection to the server established");
	return ERR_OK;
}
void BtClient::Free()
{
	TTMPMessageDestroy(iMessage);
	iMessage = NULL;
	TTMPConnectionDestroy(iConnection);
  	iConnection = NULL;
	DEBUG_LOG("Connection to the server closed");
}
//in this case -1 should be returned if communication with a server fails
#undef TTMP_ERROR
#define TTMP_ERROR -1
int BtClient::Init( char* aAddr, int aChannel, int aProfile, int aOptions, char* aPin )
{
    int handle;
    int cmdid;
    TTMPMessage msg;

    if (!aAddr)
        return -1;

    DEBUG_LOG5("BtClient::Init(%s, %d, %d, %d, %s).", aAddr, aChannel, aProfile, aOptions, aPin?aPin:"0000");

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_INIT);
    TTMP_SETFIELDSTR(iMessage, CMD_PAR_ADDRESS, aAddr);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_CHANNEL, aChannel);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_PROFILE, RemapProfile(aProfile));
    TTMP_SETFIELDINT(iMessage, CMD_PAR_OPTION, aOptions);
    if (aPin) {
        TTMP_SETFIELDSTR(iMessage, CMD_PAR_PIN, aPin);
    } else {
        TTMP_SETFIELDSTR(iMessage, CMD_PAR_PIN, BTCL_DEFAULT_PIN);
    }
    //Send message
    TTMP_SEND(iConnection, iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;
    //Parse response
    TTMP_GETFIELDINT(msg, CMD_PAR_HANDLE, &handle);
    TTMP_DESTROY(msg);
    DEBUG_LOG1("BtClient::Init() returned handle %d", handle);
    //Empty cache to prevent getting old data
    INVALIDATE_STATECACHE(handle);
    INVALIDATE_ERRORCACHE(handle);

    return handle;
}
#undef TTMP_ERROR
#define TTMP_ERROR ERR_FAILED

int BtClient::StartUsingChannel(int aHandle, const char* aData )
{
    int result;
    int cmdid;
    TTMPMessage msg;

    if ((aHandle < 0) || (aHandle >= BT_MAX_HANDLES) || !aData)
        return ERR_FAILED;

    DEBUG_LOG2("BtClient::StartUsingChannel(%d, %s).", aHandle, aData);

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_STARTUSINGCHANNEL);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_HANDLE, aHandle);
    TTMP_SETFIELDSTR(iMessage, CMD_PAR_DATA, aData);
    //Send message
    TTMP_SEND(iConnection, iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;
    //Parse response
    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
    result = RemapError(result);
    TTMP_DESTROY(msg);
    //Empty cache to prevent getting old data
    INVALIDATE_STATECACHE(aHandle);
    INVALIDATE_ERRORCACHE(aHandle);

    DEBUG_LOG1("BtClient::StartUsingChannel. Returned %d.", result);

    return result;
}

int BtClient::StopUsingChannel(int aHandle)
{
    int result;
    int cmdid;
    TTMPMessage msg;

    if ((aHandle < 0) || (aHandle >= BT_MAX_HANDLES))
        return ERR_FAILED;

    DEBUG_LOG1("BtClient::StopUsingChannel(%d).", aHandle);

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_STOPUSINGCHANNEL);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_HANDLE, aHandle);
    //Send message
    TTMP_SEND(iConnection, iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;
    //Parse response
    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
    result = RemapError(result);
    TTMP_DESTROY(msg);
    //Empty cache to prevent getting old data
    INVALIDATE_STATECACHE(aHandle);
    INVALIDATE_ERRORCACHE(aHandle);

    DEBUG_LOG1("BtClient::StopUsingChannel. Returned %d.", result);

    return result;
}

int BtClient::Release(int aHandle)
{
    int result;
    int cmdid;
    TTMPMessage msg;

    if ((aHandle < 0) || (aHandle >= BT_MAX_HANDLES))
        return ERR_FAILED;

    DEBUG_LOG1("BtClient::Release(%d).", aHandle);

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_RELEASE);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_HANDLE, aHandle);
    //Send message
    TTMP_SEND(iConnection, iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;
    //Parse response
    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
    result = RemapError(result);
    TTMP_DESTROY(msg);
    //Empty cache to prevent getting old data
    INVALIDATE_STATECACHE(aHandle);
    INVALIDATE_ERRORCACHE(aHandle);

    DEBUG_LOG1("BtClient::Release. Returned %d.", result);

    return result;
}

int BtClient::Close(int aHandle, int aForced)
{
    int result;
    int cmdid;
    TTMPMessage msg;

    if ((aHandle < 0) || (aHandle >= BT_MAX_HANDLES))
        return ERR_FAILED;

    DEBUG_LOG1("BtClient::Close(%d).", aHandle);

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_CLOSE);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_HANDLE, aHandle);
    //Send message
    TTMP_SEND(iConnection, iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;
    //Parse response
    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
    result = RemapError(result);
    TTMP_DESTROY(msg);
    //Empty cache to prevent getting old data
    INVALIDATE_STATECACHE(aHandle);
    INVALIDATE_ERRORCACHE(aHandle);

    DEBUG_LOG1("BtClient::Close. Returned %d.", result);

    return result;
}

int BtClient::Send(int aHandle, char* aBuffer, int aSize )
{
    int result;
    int cmdid;
    TTMPMessage msg;

    if ((aHandle < 0) || (aHandle >= BT_MAX_HANDLES) || !aBuffer || (aSize <= 0))
        return ERR_FAILED;

    // DEBUG_LOG1("BtClient::Send(%d).", aHandle);

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_SEND);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_HANDLE, aHandle);
    TTMP_SETFIELDBIN(iMessage, CMD_PAR_DATA, aBuffer, aSize);
    //Send message
    TTMP_SEND(iConnection, iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;
    //Parse response
    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
    result = RemapError(result);
    TTMP_DESTROY(msg);
    //Empty cache to prevent getting old data
    INVALIDATE_STATECACHE( aHandle );
    INVALIDATE_ERRORCACHE( aHandle );

    // DEBUG_LOG1("BtClient::Send. Returned %d.", result);

    return result;
}

int BtClient::Receive(int aHandle, char* aBuffer, int* aSize)
{
    if ((aHandle < 0) || (aHandle >= BT_MAX_HANDLES) || !aBuffer || (*aSize <= 0))
        return ERR_FAILED;

    TTicks now = OSALTickCount();
    if ( ( now - lastUpdatedReceive[ aHandle ] ) > BTCL_UPDATE_INTERVAL || ( now - lastUpdatedReceive[ aHandle ] ) < 0 || receiveCache[ aHandle ] < 0 )
    {
        int result, size;
        int cmdid;
        TTMPMessage msg;

        iCmdId = iCmdId++;
        TTMP_CLEAR(iMessage);
        //Prepare the command
        TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
        TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_RECEIVE);
        TTMP_SETFIELDINT(iMessage, CMD_PAR_HANDLE, aHandle);
        TTMP_SETFIELDINT(iMessage, CMD_PAR_SIZE, *aSize);
        //Send message
        TTMP_SEND(iConnection, iMessage);
retry:
        //Receive response
        TTMP_RECEIVE(iConnection, &msg);
        //Flow control
        TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
        if (cmdid < iCmdId)
            goto retry;
        //Parse response
        TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
        result = RemapError(result);
        if (result == ERR_OK) {
            TTMP_GETFIELDINT(msg, CMD_PAR_SIZE, &size);
            if (size) {
                TTMP_GETFIELDSIZE(msg, CMD_PAR_DATA, &size);
                TTMP_GETFIELDBIN(msg, CMD_PAR_DATA, aBuffer, size);
            }
            *aSize = size;
        } else
            *aSize = 0;
        TTMP_DESTROY(msg);

        if ( *aSize == 0 ) lastUpdatedReceive[ aHandle ] = now;
        receiveCache[ aHandle ] = result;

        //Empty cache to prevent getting old data
        INVALIDATE_STATECACHE(aHandle);
        INVALIDATE_ERRORCACHE(aHandle);

        return result;
    }
    else
    {
        *aSize = 0;
        return receiveCache[ aHandle ];
    }
}

int BtClient::GetInternalState( int aHandle )
{
    if ((aHandle < 0) || (aHandle >= BT_MAX_HANDLES))
        return -1;

    TTicks now = OSALTickCount();
    if ( ( now - lastUpdatedState[ aHandle ] ) > BTCL_UPDATE_INTERVAL || ( now - lastUpdatedState[ aHandle ] ) < 0 || stateCache[ aHandle ] < 0 )
    {
        int state;
        int cmdid;
        TTMPMessage msg;

        // DEBUG_LOG1("BtClient::GetInternalState(%d).", aHandle);

        iCmdId = iCmdId++;
        TTMP_CLEAR(iMessage);
        //Prepare the command
        TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
        TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_GETSTATE);
        TTMP_SETFIELDINT(iMessage, CMD_PAR_HANDLE, aHandle);
        //Send message
        TTMP_SEND(iConnection, iMessage);
retry:
        //Receive response
        TTMP_RECEIVE(iConnection, &msg);
        //Flow control
        TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
        if (cmdid < iCmdId)
            goto retry;
        //Parse response
        TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &state);
        //RemapState(state);
        TTMP_DESTROY(msg);
        //Empty cache to prevent getting old data
        INVALIDATE_STATECACHE(aHandle);
        INVALIDATE_ERRORCACHE(aHandle);

        lastUpdatedState[ aHandle ] = now;
        stateCache[ aHandle ] = state;

        // DEBUG_LOG1("BtClient::GetInternalState. Returned %d.", state);

        return state;
    }
    else
        return stateCache[ aHandle ];
}

int BtClient::GetState( int aHandle )
{
    return RemapState(GetInternalState(aHandle));
}

int BtClient::GetError( int aHandle )
{
    if ((aHandle < 0) || (aHandle >= BT_MAX_HANDLES))
        return ERR_FAILED;

    TTicks now = OSALTickCount();
    if ( ( now - lastUpdatedError[ aHandle ] ) > BTCL_UPDATE_INTERVAL || ( now - lastUpdatedError[ aHandle ] ) < 0 || errorCache[ aHandle ] < 0 )
    {
        int result;
        int cmdid;
        TTMPMessage msg;

        // DEBUG_LOG1("BtClient::GetError(%d).", aHandle);

        iCmdId = iCmdId++;
        TTMP_CLEAR(iMessage);
        //Prepare the command
        TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
        TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_GETERROR);
        TTMP_SETFIELDINT(iMessage, CMD_PAR_HANDLE, aHandle);
        //Send message
        TTMP_SEND(iConnection, iMessage);
retry:
        //Receive response
        TTMP_RECEIVE(iConnection, &msg);
        //Flow control
        TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
        if (cmdid < iCmdId)
            goto retry;
        //Parse response
        TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
        result = RemapError(result);
        TTMP_DESTROY(msg);
        //Empty cache to prevent getting old data
        INVALIDATE_STATECACHE(aHandle);
        INVALIDATE_ERRORCACHE(aHandle);

        lastUpdatedError[ aHandle ] = now;
        errorCache[ aHandle ] = result;

        // DEBUG_LOG1("BtClient::GetError. Returned %d.", result);

        return result;
    }
    else
        return errorCache[ aHandle ];
}

int BtClient::SetFlags( int aHandle, int aOpts )
{
    int result;
    int cmdid;
    TTMPMessage msg;

    if ((aHandle < 0) || (aHandle >= BT_MAX_HANDLES))
        return ERR_FAILED;

    DEBUG_LOG1("BtClient::SetFlags(%d).", aHandle);

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_SETFLAGS);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_HANDLE, aHandle);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_OPTION, aOpts);
    //Send message
    TTMP_SEND(iConnection, iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;
    //Parse response
    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
    result = RemapError(result);
    TTMP_DESTROY(msg);

    DEBUG_LOG("BtClient::SetFlags. Returned.");

    return result;
}

int BtClient::GetFlags( int aHandle )
{
    int result;
    int cmdid;
    TTMPMessage msg;

    if ((aHandle < 0) || (aHandle >= BT_MAX_HANDLES))
        return -1;

    DEBUG_LOG1("BtClient::GetFlags(%d).", aHandle);

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_GETFLAGS);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_HANDLE, aHandle);
    //Send message
    TTMP_SEND(iConnection, iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;
    //Parse response
    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
    TTMP_DESTROY(msg);

    DEBUG_LOG1("BtClient::GetFlags. Returned %d.", result);

    return result;
}

int BtClient::IsProfileActive( int aProfile )
{
    int result, profile;
    int cmdid;
    TTMPMessage msg;

    DEBUG_LOG1("BtClient::IsProfileActive(%d).", aProfile);

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_ISPROFILEACTIVE);
    profile = RemapError(aProfile);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_PROFILE, profile);
    //Send message
    TTMP_SEND(iConnection, iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;
    //Parse response
    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
    TTMP_DESTROY(msg);

    DEBUG_LOG1("BtClient::GetFlags. Returned %d.", result);

    return result;
}
//in this case -1 should be returned if communication with a server fails
#undef TTMP_ERROR
#define TTMP_ERROR -1
int BtClient::RequestConnection()
{
    int handle;
    int cmdid;
    TTMPMessage msg;

    DEBUG_LOG("BtClient::RequestConnection().");

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_REQUESTCONNECTION);
    //Send message
    TTMP_SEND(iConnection, iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;
    //Parse response
    TTMP_GETFIELDINT(msg, CMD_PAR_HANDLE, &handle);
    TTMP_DESTROY(msg);
    DEBUG_LOG1("BtClient::RequestConnection() returned handle %d", handle);
    //Empty cache to prevent getting old data
    INVALIDATE_STATECACHE(handle);
    INVALIDATE_ERRORCACHE(handle);

    return handle;
}
int BtClient::IsConnectionAllowed( int aHandle )
{

    if ((aHandle < 0) || (aHandle >= BT_MAX_HANDLES))
        return 1;

    // DEBUG_LOG1("BtClient::IsConnectionAllowed(%d).", aHandle);

    return (GetInternalState(aHandle) == BT_STATE_CONNECTING) ? 1 : 0;
}
//in this case -1 should be returned if communication with a server fails
#undef TTMP_ERROR
#define TTMP_ERROR
void BtClient::ReleaseConnectionRequest( int aHandle )
{
    int cmdid;
    TTMPMessage msg;

    if ((aHandle < 0) || (aHandle >= BT_MAX_HANDLES))
        return;

    DEBUG_LOG1("BtClient::ReleaseConnectionRequest(%d).", aHandle);

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_RELEASECONNECTIONREQUEST);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_HANDLE, aHandle);
    //Send message
    TTMP_SEND(iConnection, iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;
    TTMP_DESTROY(msg);
    //Empty cache to prevent getting old data
    INVALIDATE_STATECACHE(aHandle);
    INVALIDATE_ERRORCACHE(aHandle);

    DEBUG_LOG("BtClient::ReleaseConnectionRequest. Returned.");

    return;
}

// This forces the connected device to be re-connected. This is needed when devices are
// in an error state and needs to be reconnected in order to work again.
#undef TTMP_ERROR
#define TTMP_ERROR -1
int BtClient::Reconnect( int aHandle )
{
    int result = ERR_OK;
  int cmdid;
    TTMPMessage msg;

    if ((aHandle < 0) || (aHandle >= BT_MAX_HANDLES))
        return ERR_FAILED;

    DEBUG_LOG1("BtClient::Reconnect(%d).", aHandle);

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_RECONNECT);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_HANDLE, aHandle);
    //Send message
    TTMP_SEND(iConnection, iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;
    // Parse response 
    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);

    TTMP_DESTROY(msg);
    DEBUG_LOG1("BtClient::Reconnect. Returned %d.", result);
    return result;
}

int BtClient::ExitSniffMode( char* bdaddr )
{
  int result = ERR_OK;
  int cmdid;
    TTMPMessage msg;
    if (!bdaddr)
    return -1;
    DEBUG_LOG1("BtClient::ExitSniffMode(%s).", bdaddr);
    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_EXIT_SNIFF_MODE);
    TTMP_SETFIELDSTR(iMessage, CMD_PAR_ADDRESS, bdaddr);
    //Send message
    TTMP_SEND(iConnection, iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;
    // Parse response 
    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);

    TTMP_DESTROY(msg);
    DEBUG_LOG1("BtClient::ExitSniffMode. Returned %d.", result);
    return result;
}

#undef TTMP_ERROR
#define TTMP_ERROR -1

int BtClient::GetBTConnectionInfo(int aHandle, char* aAddr, int aSize)
{
    int result = ERR_OK;
    int cmdid;

    TTMPMessage msg;
    if ((aHandle < 0) || (aHandle >= BT_MAX_HANDLES))
        return ERR_FAILED;

    DEBUG_LOG2("BtClient::GetBTConnectionInfo(%d) size(%d).", aHandle, aSize);

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_GETBTCONNECTIONINFO);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_HANDLE, aHandle);
    //Send message
    TTMP_SEND(iConnection, iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
    {
        DEBUG_LOG2("BtClient::GetBTConnectionInfo retry cmdid=%d iCmdId=%d.", cmdid, iCmdId);
    goto retry;
    }
    //Parse response
    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
    result = RemapError(result);
    if (result == ERR_OK) 
    {
        DEBUG_LOG("BtClient::GetBTConnectionInfo result OK.");
        TTMP_GETFIELDSTR(msg, CMD_PAR_ADDRESS, aAddr, aSize);
        DEBUG_LOG1("BtClient::GetBTConnectionInfo. [addr=%s]", aAddr);
    } 
    else
    {
        DEBUG_LOG("BtClient::GetBTConnectionInfo result FAILED.");
        result = ERR_FAILED;
    }
    TTMP_DESTROY(msg);
    return result;
}

// get local bt name
int BtClient::GetBTName( char *aName, int aSize )
{
    int cmdid, result = ERR_OK, size;
    TTMPMessage msg;

    if (!aName)
        return ERR_FAILED;

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_GETBTNAME);
    //Send message
    TTMP_SEND(iConnection, iMessage);
    TTMP_CLEAR(iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;
    //Parse response
    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
    result = RemapError(result);
    if (result == ERR_OK) {
        TTMP_GETFIELDINT(msg, CMD_PAR_SIZE, &size);
        if (size && size <= aSize) {
            TTMP_GETFIELDSIZE(msg, CMD_PAR_DATA, &size);
            TTMP_GETFIELDBIN(msg, CMD_PAR_DATA, aName, size);
        } else
            result = ERR_FAILED;
    } else
        result = ERR_FAILED;
    TTMP_DESTROY(msg);

    DEBUG_LOG1("BtClient::GetBTName. Returned %d.", result);

    return result;
}

int BtClient::SetBTName( char *aName )
{
    int result = ERR_OK;
    int cmdid;
    TTMPMessage msg;

    if (!aName)
        return ERR_FAILED;

    DEBUG_LOG1("BtClient::SetBTName(%s).", aName);

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_SETBTNAME);
    TTMP_SETFIELDSTR(iMessage, CMD_PAR_DATA, aName);
    //Send message
    TTMP_SEND(iConnection, iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;
    //Parse response
    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
    result = RemapError(result);
    TTMP_DESTROY(msg);

    DEBUG_LOG1("BtClient::SetBTName. Returned %d.", result);

    return result;
}

#undef TTMP_ERROR
#define TTMP_ERROR ERR_FAILED

int BtClient::GetHidRequest(char *aAddress, int aSize)
{
  int result = ERR_FAILED;
  if (iMessage != NULL)
  {
    int cmdid;
    TTMPMessage msg;
    iCmdId++;
    TTMP_CLEAR(iMessage);
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_REQUEST_HID_CONN);
    TTMP_SEND(iConnection, iMessage);

retry:
    TTMP_RECEIVE(iConnection, &msg);
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;


    TTMP_GETFIELDBIN(msg, CMD_PAR_DATA, aAddress, aSize);
    if (!aAddress)
        return ERR_FAILED;

    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
    result = RemapError(result);

    TTMP_DESTROY(msg);
  }
  return result;
}

int BtClient::ConnectToHid(char *aAddress, int size)
{
    int result = ERR_FAILED;
    int cmdid;
    TTMPMessage msg;

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_CONNECT_HID);
    TTMP_SETFIELDSTR(iMessage, CMD_PAR_DATA, aAddress);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_SIZE, size);

    DEBUG_LOG("BtClient::ConnectToHid. send" );
    TTMP_SEND(iConnection, iMessage);

retry:
    TTMP_RECEIVE(iConnection, &msg);
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;

    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
    result = RemapError(result);

    DEBUG_LOG1("BtClient::ConnectToHid. Returned %d.", result);
    TTMP_DESTROY(msg);
    return result;
}

int BtClient::StartHid()
{
    int result = ERR_FAILED;
    int cmdid;
    TTMPMessage msg;

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_START_HID);

    TTMP_SEND(iConnection, iMessage);

retry:
    TTMP_RECEIVE(iConnection, &msg);
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
      goto retry;

    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
    result = RemapError(result);

    DEBUG_LOG1("BtClient::StartHid Returned %d.", result);
    TTMP_DESTROY(msg);
    return result;
}


int BtClient::DisconnectHid()
{
    int result = ERR_FAILED;
    int cmdid;
    TTMPMessage msg;

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_DISCONNECT_HID);

    TTMP_SEND(iConnection, iMessage);

retry:
    TTMP_RECEIVE(iConnection, &msg);
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;

    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
    result = RemapError(result);

    DEBUG_LOG1("BtClient::DisconnectHid. Returned %d.", result);
    TTMP_DESTROY(msg);
    return result;
}

/**
 * retreives the actual MTU sizes for the given connection
 * returns -1 on failure, 0 on success
 */
int BtClient::GetMtuSize( int aHandle, int * imtu, int * omtu )
{
    int result=0;
    int cmdid;
    TTMPMessage msg;

    if ((aHandle < 0) || (aHandle >= BT_MAX_HANDLES))
        return -1;

    DEBUG_LOG1("BtClient::GetMtuSize(%d).", aHandle);

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_GET_MTU_SIZE);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_HANDLE, aHandle);
    //Send message
    TTMP_SEND(iConnection, iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;
    //Parse response
    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
    TTMP_GETFIELDINT(msg, CMD_PAR_DATA, imtu);
    TTMP_GETFIELDINT(msg, CMD_PAR_SIZE, omtu);
    TTMP_DESTROY(msg);

    DEBUG_LOG3("BtClient::GetMtuSize. Returned result=%d. imtu=%d, omtu=%d.", result, *imtu, *omtu);

    return result;
}

/**
 * request for a /dev/rfcommx device coupled to the given bdaddr
 *
 * returns -1 on failure, 0 on success
 */
int BtClient::CreateTTY( int aHandle, int aTtyId)
{
    int result=0;
    int cmdid;
    TTMPMessage msg;

    DEBUG_LOG2("%d, %d", aHandle, aTtyId);

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_CREATE_TTY);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_HANDLE, aHandle);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_DATA, aTtyId);
    //Send message
    TTMP_SEND(iConnection, iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;
    //Parse response
    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
    TTMP_DESTROY(msg);
    DEBUG_LOG1("result %d", result);
    return result;
}

/**
 * release the /dev/rfcommx device coupled to this btclient
 *
 * returns -1 on failure, 0 on success
 */
int BtClient::ReleaseTTY( int aHandle)
{
    int result=0;
    int cmdid;
    TTMPMessage msg;

    DEBUG_LOG1("%d", aHandle);

    iCmdId = iCmdId++;
    TTMP_CLEAR(iMessage);
    //Prepare the command
    TTMP_SETFIELDINT(iMessage, CMD_ID, iCmdId);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_COMMAND, CMD_RELEASE_TTY);
    TTMP_SETFIELDINT(iMessage, CMD_PAR_HANDLE, aHandle);
    //Send message
    TTMP_SEND(iConnection, iMessage);
retry:
    //Receive response
    TTMP_RECEIVE(iConnection, &msg);
    //Flow control
    TTMP_GETFIELDINT(msg, CMD_ID, &cmdid);
    if (cmdid < iCmdId)
        goto retry;
    //Parse response
    TTMP_GETFIELDINT(msg, CMD_PAR_RESULT, &result);
    TTMP_DESTROY(msg);
    DEBUG_LOG1("result %d", result);
    return result;
}

#undef TTMP_ERROR
